from functools import total_ordering

from coalib.misc.Decorators import generate_repr
from coalib.results.SourcePosition import SourcePosition


@generate_repr("start", "end")
@total_ordering
class SourceRange:
    def __init__(self, start, end=None):
        """
        Creates a new SourceRange.

        :param start: A SourcePosition indicating the start of the range.
        :param end:   A SourcePosition indicating the end of the range. If
                      `None` is given, the start object will be used here. end
                      must be in the same file and be greater than start as
                      negative ranges are not allowed.
        """
        assert isinstance(start, SourcePosition)

        self._start = start
        if end is None:
            self._end = self.start
        else:
            assert isinstance(end, SourcePosition)
            assert self.start.file == end.file
            assert end >= self.start

            self._end = end

    @classmethod
    def from_values(cls,
                    file,
                    start_line=None,
                    start_column=None,
                    end_line=None,
                    end_column=None):
        start = SourcePosition(file, start_line, start_column)
        if not end_line:
            end = None
        else:
            end = SourcePosition(file, end_line, end_column)

        return cls(start, end)

    @property
    def file(self):
        return self.start.file

    @property
    def start(self):
        return self._start

    @property
    def end(self):
        return self._end

    def __eq__(self, other):
        if not isinstance(other, SourceRange):
            return False

        return self.start == other.start and self.end == other.end

    def __lt__(self, other):
        if not isinstance(other, SourceRange):
            raise TypeError("Cannot compare SourceRange with a variable of"
                            "another type")

        if self.start != other.start:
            return self.start < other.start

        return self.end < other.end



import difflib
from pyprint.ConsolePrinter import ConsolePrinter

from coalib.results.Diff import Diff
from coalib.results.result_actions.ResultAction import ResultAction
from coalib.results.Result import Result


def format_line(line, real_nr="", sign="|", mod_nr="", symbol="", ):
    return "|{:>4}{}{:>4}|{:1}{}".format(real_nr,
                                         sign,
                                         mod_nr,
                                         symbol,
                                         line.rstrip("\n"))


def print_beautified_diff(difflines, printer):
    current_line_added = None
    current_line_subtracted = None
    for line in difflines:
        if line.startswith("@@"):
            values = line[line.find("-"):line.rfind(" ")]
            subtracted, added = tuple(values.split(" "))
            current_line_added = int(added.split(",")[0][1:])
            current_line_subtracted = int(subtracted.split(",")[0][1:])
        elif line.startswith("---"):
            printer.print(format_line(line[4:], real_nr="----"), color="red")
        elif line.startswith("+++"):
            printer.print(format_line(line[4:], mod_nr="++++"),
                          color="green")
        elif line.startswith("+"):
            printer.print(format_line(line[1:],
                                      mod_nr=current_line_added,
                                      symbol="+"),
                          color="green")
            current_line_added += 1
        elif line.startswith("-"):
            printer.print(format_line(line[1:],
                                      real_nr=current_line_subtracted,
                                      symbol="-"),
                          color="red")
            current_line_subtracted += 1
        else:
            printer.print(format_line(line[1:],
                                      real_nr=current_line_subtracted,
                                      mod_nr=current_line_added,
                                      symbol=" "))
            current_line_subtracted += 1
            current_line_added += 1


class ShowPatchAction(ResultAction):
    @classmethod
    def is_applicable(cls, result):
        return isinstance(result, Result) and result.diffs is not None

    def apply(self,
              result,
              original_file_dict,
              file_diff_dict,
              colored: bool=True):
        '''
        Print a diff of the patch that would be applied.

        :param colored: Wether or not to use colored output.
        '''
        printer = ConsolePrinter(colored)

        for filename, this_diff in sorted(result.diffs.items()):
            original_file = original_file_dict[filename]
            diff = file_diff_dict.get(filename, Diff())
            current_file = diff.apply(original_file)
            new_file = this_diff.apply(current_file)
            print_beautified_diff(difflib.unified_diff(current_file,
                                                       new_file,
                                                       fromfile=filename,
                                                       tofile=filename),
                                   printer)

        return file_diff_dict



from functools import total_ordering

from coalib.misc.Decorators import generate_repr


@generate_repr("file", "line", "column")
@total_ordering
class SourcePosition:
    def __init__(self, file, line=None, column=None):
        """
        Creates a new result position object that represents the position of a
        result in the source code.

        :param file:            The filename or None.
        :param line:            The line in file or None, the first line is 1.
        :param column:          The column indicating the character. The first
                                one in a line is 1.
        :raises AssertionError: If a line number without a file is provided.
        """
        assert isinstance(file, str), "file must be a string!"
        assert isinstance(line, int) or not line, "line must be an int!"
        assert isinstance(column, int) or not column, "column must be an int!"
        assert file is not None or line is None, ("A line must be associated "
                                                  "to a file.")
        assert line is not None or column is None, ("A column can only be set "
                                                    "if a line is set.")

        self._file = file
        self._line = line
        self._column = column

    @property
    def file(self):
        return self._file

    @property
    def line(self):
        return self._line

    @property
    def column(self):
        return self._column

    def __str__(self):
        return "file: {}, line: {}, column: {}".format(
            str(repr(self.file)),
            str(self.line),
            str(self.column))

    def __eq__(self, other):
        return (self.file == other.file and
                self.line == other.line and
                self.column == other.column)

    def __lt__(self, other):
        # Now either both file members are None or both are set
        if self.file != other.file:
            return self.file < other.file

        # Show results with a no or lesser line number first
        if (self.line is None) != (other.line is None):
            return self.line is None

        if self.line != other.line:
            return self.line < other.line

        # Show results with no or lesser column first
        if (self.column is None) != (other.column is None):
            return self.column is None

        if self.column != other.column:
            return self.column < other.column

        return False
